(ns clojureDemo.OpenCv)
  
(import '(java.awt Rectangle Robot Toolkit) )
(import '(java.awt.image BufferedImage) )
(import '(java.awt Graphics2D Panel))
(import '(java.io File) )
(import '(javax.imageio ImageIO) )
(import '(javax.swing JFrame))
(import '(org.apache.commons.io FileUtils))
(import clojure.lang.LazySeq)
(import '(name.audet.samuel.javacv.jna highgui cv cxcore 
				       cxcore$IplImage highgui$CvCapture$PointerByReference  
				       highgui$CvVideoWriter$PointerByReference cxcore$IplImage$PointerByReference))
(import '(name.audet.samuel.javacv CanvasFrame JavaCvErrorCallback))

(.redirectError (JavaCvErrorCallback.))

(use 'clojure.contrib.repl-utils)
;(use 'clojureDemo.Defines)
;(use '[clojureDemo.Xuggle :only (cache)])

;this is still a work in progress, I'll come back to it later when I understand 
;jna more thoroughly.  the important abstraction here is 
;video-seq, which gives a lazy sequence of Intel Image Processing library images.

(defn naturals [] (iterate inc 0))

(defn- makePanel [image] (proxy [Panel] [] (paint [g]  (.drawImage g image 0 0 nil))))


(defmulti  display "Creates a JFrame and displays a buffered image"  class)

(defmethod display 
  BufferedImage  [image] 
  (let [panel (makePanel image)
	frame (JFrame. "Oh Yeah!")]
    (.add frame panel) 
    (.pack frame) 
    (.setVisible frame true ) 
    (.setSize frame(.getWidth image) (.getHeight image))))

(defmethod display
  cxcore$IplImage [image]
  ( display (.getBufferedImage image)))
  
(defmethod display
  String [image]
  (display (highgui/cvLoadImage image highgui/CV_LOAD_IMAGE_COLOR)))

(defmethod display
  LazySeq [s]
  (display (first s)))



(def ext "jpg") 
;see below for the rationale for this choice of extention.

(def cache-location "/home/r/Desktop/vision-cache/")

(defn close-capture
  [capture]
  (highgui/cvReleaseCapture (highgui$CvCapture$PointerByReference. capture)))

(defn close-writer
  [writer] (highgui/cvReleaseVideoWriter (highgui$CvVideoWriter$PointerByReference. writer))) 

(defn- cache-path
  [video]
  (File. cache-location (.getName video)))

(defn- already-cached
  "this is the simplest and most retarded way to do it"
  [video]
  (.exists (cache-path video)))

(defn write-frame 
  [capture target-dir n]
  (let [image (highgui/cvQueryFrame capture)]
    (if (nil? image) false
	(highgui/cvSaveImage (str (File. target-dir (str n "."  ext))) image))))

(defn- write-frame-bad
  [capture target-dir n]
  (println (str "saving frame: " n))
  (let [image (highgui/cvQueryFrame capture)]
    (if (nil? image) false
	( ImageIO/write (.getBufferedImage image) ext (File. target-dir (str n "."  ext))))))

(defn- write-frames
  [video target-dir]
  (let [capture (highgui/cvCreateFileCapture (.getPath video))]
   (dorun 
    (for [n (naturals) :while (write-frame capture target-dir n) ] nil ))
    (highgui/cvReleaseCapture (highgui$CvCapture$PointerByReference. capture))))

(defn- cache-frames
  [cache-location video]
  (time 
   (do
     (println "\"caching entire video structure... this will take a while... go get a snack or something :)\"")
     (FileUtils/deleteDirectory (cache-path video))
     (FileUtils/forceMkdir (cache-path video))
     (write-frames video  (cache-path video)))))

;(defn cache
;  [video]
;  (if (already-cached video) nil (cache-frames cache-location video)))
 
(defn video-len 
  [video]
  (alength (.list (cache-path video))))
(def video-len (memoize video-len))
 
(defn video-data
  "since the opencv version is so absolutely unreliable..."
  [video]
  (let
      [capture (highgui/cvCreateFileCapture (.getPath video))
       info {:length      (video-len video)
	     :width       (highgui/cvGetCaptureProperty capture highgui/CV_CAP_PROP_FRAME_WIDTH)
	     :height      (highgui/cvGetCaptureProperty capture highgui/CV_CAP_PROP_FRAME_HEIGHT)
	     :fps         (highgui/cvGetCaptureProperty capture highgui/CV_CAP_PROP_FPS)
	     :codec       (highgui/cvGetCaptureProperty capture highgui/CV_CAP_PROP_FOURCC)}]
    (close-capture capture)
    info))
(def video-data (memoize video-data))

(defn video-frame-path
     [video n]
     (File. (cache-path video)  (str n "."  ext)))


(defn- video-frame-ipl
  [video n] 
;  (cache video)
  (let 
      [c++-managed  (highgui/cvLoadImage (str (File. (cache-path video) (str n "."  ext))) highgui/CV_LOAD_IMAGE_COLOR)
       jvm-managed (.clone c++-managed)] 
	 ;this bit with the cloning is so I can deal with Garbage Collection once and for all.
	 ;the cpp-managed image must be manually Garbage Collected, but it's clone is managed by 
         ;the JVM's Garbage Collector.  By getting rid of the c++ part right here and now, no 
         ;other function has to worry about manual garbage collection ever again.
         ;Unfortunately, this doesn't seem to work for certain types of files. It's not file-size
         ;which is the issue, but something involving the image header.
    (cxcore/cvReleaseImage (.pointerByReference c++-managed))
    jvm-managed
))
  

(defn- video-frame-buffered
  "takes one frame from a video in constant time"
  [video n]
 ; (cache video)
  (ImageIO/read (File. (cache-path video) (str n "."  ext))))

(defn video-frame [video n] (video-frame-buffered video n))

(defn- dumb-write 
  [video n writer]
  (let 
      [c++-managed  (highgui/cvLoadImage (str (File. (cache-path video) (str n ext))) highgui/CV_LOAD_IMAGE_COLOR)]
    (highgui/cvWriteFrame writer  c++-managed)
    (cxcore/cvReleaseImage (cxcore$IplImage$PointerByReference. c++-managed))))

(defn video-seq
  "makes a lazy sequence of IPL images"
  ;additionally, I want to pass metadata around with the sequence.
  [video] ;(cache video)
  (map #(video-frame video %) (range (video-len video))))
(defn video-writer
  "uses data about the video to make a writer"
  [data fileTarget]
  (highgui/cvCreateVideoWriter 
   (str fileTarget)
   
   ;(highgui/CV_FOURCC \P,\I,\M,\1) ;= MPEG-1 codec  (112913.386195 msecs) (104 MB)
   ;(highgui/CV_FOURCC \M,\J,\P,\G) ;= motion-jpeg codec (crashed)
   ;(highgui/CV_FOURCC \M,\P,\4,\2) ;= MPEG-4.2 codec (107184.186774 msecs) (82 MB)
   ;(highgui/CV_FOURCC \D,\I,\V,\3) ;= MPEG-4.3 codec (118308.933328 msecs)  (83 MB)
   (highgui/CV_FOURCC \D,\I,\V,\X) ;= MPEG-4 codec    (99037.738131 msecs)  (85 MB)
   ;(highgui/CV_FOURCC \U,\2,\6,\3) ;= H263 codec     (101141.993551 msecs) (89 MB)
   ;(highgui/CV_FOURCC \I,\2,\6,\3) ;= H263I codec      (crashed)
   ;(highgui/CV_FOURCC \F,\L,\V,\1) ;= FLV1 codec     (104307.567802 msecs) (93 MB)
   ;(:codec data)                    ;= whatever the movie originally had.  (98278.694169 msecs) (1.9 GB)  

   (:fps data) (cxcore/cvSize (:width data) (:height data)) 
   1; 1 here means that we're writing in color.
    ; you cannot change it to 0 to write in 
    ; black and white. Everything just crashes instead.
    ; what a useful paramater.
   ))


(defn naturals []  (iterate inc 0))


(defn write-frame-2
  [writer frame]
  (let [c++-frame (cxcore$IplImage/createFrom frame)]
    (highgui/cvWriteFrame writer c++-frame)
   ; (cxcore/cvReleaseImage (.pointerByReference c++-frame)))
)
    frame)

(defn save-seq 
  [writer video-seq]
  (map #(write-frame-2 writer %) video-seq))

(defmacro trans-save
"there's a small problem with trans-save --- it IS
truly transitive, but it does too much work....
sometimes it writes files twice.
this is functionally correct though."
  [target config video-seq]
  `(let [writer# (video-writer ~config ~target)]
     (do
       (dorun (save-seq writer# ~video-seq))
       (close-writer writer#)
       ~video-seq)))



(comment 
(do (use :reload-all 'clojureDemo.OpenCv) (in-ns 'clojureDemo.OpenCv))
)
